import os
from pypdf import PdfReader

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = CURRENT_DIR.split('src')[0]
CV_DIR = os.path.join(ROOT_DIR, 'data', 'cvs')


class Dataloader:
    def __init__(self, filepath=CV_DIR):
        self.filepath = filepath

    @staticmethod
    def _load_pdf(filepath):
        reader = PdfReader(filepath)

        return ' '.join([page.extract_text() for page in reader.pages])

    def load_pdf(self, filepath: str):
        if filepath is None:
            raise ValueError("Filename cannot be None")

        # Verify if file is a pdf
        if not filepath.endswith('.pdf'):
            raise ValueError("File must be a pdf")

        # Verify if the file exists
        if not os.path.exists(filepath):
            raise FileNotFoundError(
                f"Filepath {filepath} not found in {self.filepath}")

        return Dataloader._load_pdf(filepath)

    def load_all_pdfs(self):
        # Iterate over all files in the data folder
        for file in os.listdir(self.filepath):
            if file.endswith('.pdf'):
                yield os.path.join(self.filepath, file)



from datasets import Dataset

class ManualAnnotation:
    @staticmethod
    def upload_to_hub(filepath, dataset_name="innovpoint/curriculum-vitae",config_name="evaluation"):
        Dataset.from_json(filepath, split="test").push_to_hub(dataset_name, config_name=config_name, private=True)


from cv_parsing.orms.ORM import ORM


class Education(ORM):
    def __init__(self, json_data) -> None:
        super().__init__(class_name="education", json_data=json_data)

    def evaluate(predictions: list, references: list, relaxed):
        return ORM._evaluate(predictions, references, relaxed, keys_to_align=["course", "institution"], class_type=Education)



import os
from json_repair import repair_json as rp
from cv_parsing.exceptions.JSONException import JSONException

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(CURRENT_DIR, 'schema.jsonl'), 'r', encoding='utf-8') as f:
    JSON_STR = f.read()
    SCHEMA_DICT = rp(JSON_STR, return_objects=True)

def repair_json(json_str) -> dict:
    if json_str == "":
        raise JSONException("Empty JSON string")

    # Consider all text after the first '{' as a JSON object and the last '}' as the end of the JSON object
    first_bracket = json_str.find('{')
    last_bracket = json_str.rfind('}')

    if first_bracket == -1 or last_bracket == -1:
        raise JSONException("Invalid JSON string")

    return rp(json_str[first_bracket:last_bracket+1], return_objects=True)



import re
from cv_parsing.orms.ORM import ORM
from cv_parsing.utils.metrics import NO_STR_ESCAPE


class PersonalInformation(ORM):
    def __init__(self, json_data) -> None:
        super().__init__(class_name="personal_information", json_data=json_data)

    @staticmethod
    def process_phone_numbers(data: list):
        for item in data:
            item.data['phone'] = item.data['phone'].replace('351', '')
            integers = re.findall(r'\d+', item.data['phone'])

            if len(integers) == 0:
                item.data['phone'] = NO_STR_ESCAPE
            else:
                item.data['phone'] = int(''.join(integers))

        return data

    @staticmethod
    def evaluate(predictions: list, references: list, relaxed):
        predictions = PersonalInformation.process_phone_numbers(predictions)
        references = PersonalInformation.process_phone_numbers(references)

        return ORM._evaluate(predictions, references, relaxed, keys_to_align=["name", "email"], class_type=PersonalInformation)

    def keys(self, relaxed, dict_included, schema=None):
        keys = super().keys(relaxed, dict_included, schema)

        if relaxed:
            keys.pop('phone')

        return keys



import requests
import os
import time
from cv_parsing.parser.ParsingStrategy import ParsingStrategy
from cv_parsing.orms.Person import Person
from cv_parsing.exceptions.PromptException import PromptException
from cv_parsing.exceptions.JSONException import JSONException
from cv_parsing.utils.json_utils import repair_json
import logging

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = CURRENT_DIR.split("src")[0]
DATA_DIR = os.path.join(ROOT_DIR, "data")


API_URL = "https://api-inference.huggingface.co/models"


class HuggingFaceApiStrategy(ParsingStrategy):
    def __init__(self, prompts: list, model=None, temperature=None, api_key=None) -> None:
        super().__init__("HuggingFaceApi")

        self.prompts = prompts

        self.model = model or self.env('HF_MODEL')

        self.api_key = api_key or self.env('HF_API_KEY')

        self.temperature = temperature or self.env('MODEL_TEMPERATURE')

        self.header = {"Authorization": f"Bearer {self.api_key}"}

    def _validate_model_output(self, json_response) -> dict:
        if not isinstance(json_response, list):
            raise PromptException("Invalid Hugging Face Response. Is not list")

        if len(json_response) == 0:
            raise PromptException(
                "Invalid Hugging Face Response. List is empty")

        if 'generated_text' not in json_response[0]:
            raise PromptException(
                "Invalid Hugging Face Response. Generated Text Not Present")

        try:
            json = repair_json(json_response[0]['generated_text'])
        except JSONException as e:
            raise PromptException(
                "Invalid Hugging Face Response. JSON Repair Failed")

        if not isinstance(json, dict):
            raise PromptException(
                "Invalid Hugging Face Response. JSON Repair Response is not dict")

        return json

    def parse(self, normalized_text: str):
        dict_results = {}

        for prompt in self.prompts:

            if 'Phi-3' in self.model:
                max_tokens = 4_096 - \
                    (len(prompt.substitute(cv=normalized_text).split(' '))*3)
            else:
                max_tokens = 4_096

            for _ in range(3):
                response = requests.post(
                    f"{API_URL}/{self.model}", headers=self.header, json={
                        "inputs": prompt.substitute(cv=normalized_text),
                        "parameters": {
                            'max_new_tokens': max_tokens,
                            "return_full_text": False,      
                        },
                        "options": {
                            "wait_for_model": True,
                            "use_cache": False
                        }
                    })

                if response.status_code == 200:
                    break

                if response.status_code == 422:
                    raise OverflowError(
                        f"Document too large for model {response}")

                print(
                    f"Retrying {self.model} | {response.status_code}. Sleeping 1 second.")

                time.sleep(1)

            response = response.json()

            print(f"Response {self.model} | {response}")

            try:
                dict_results.update(self._validate_model_output(response))
            except PromptException as e:
                print(f"{self.model} | {response}")
                raise PromptException(
                    f"Invalid Hugging Face Response {e} {response}")

        return Person(dict_results)

    @staticmethod
    def models():
        with open(os.path.join(DATA_DIR, "models.jsonl"), 'r', encoding='utf-8') as f:
            return repair_json(f.read())["HuggingFaceApi"]



from cleantext import clean


def clean_text(raw_text: str) -> str:
    return clean(raw_text,
          fix_unicode=True,               # fix various unicode errors
          to_ascii=False,                  # transliterate to closest ASCII representation
          lower=False,                     # lowercase text
          # fully strip line breaks as opposed to only normalizing them
          no_line_breaks=False,
          keep_two_line_breaks=False,
          strip_lines=True,
          no_urls=False,                   # replace all URLs with a special token
          no_emails=False,                 # replace all email addresses with a special token
          no_phone_numbers=False,          # replace all phone numbers with a special token
          no_numbers=False,                # replace all numbers with a special token
          no_digits=False,                # replace all digits with a special token
          no_currency_symbols=False,      # replace all currency symbols with a special token
          no_punct=False,                 # remove punctuations
          lang="en"                       # set to 'de' for German special handling
          )



from abc import ABC
from cv_parsing.utils.json_utils import SCHEMA_DICT
from cv_parsing.exceptions.ORMException import ORMException
from cv_parsing.exceptions.EvaluationException import EvaluationException
from cv_parsing.utils.logger import logging
from cv_parsing.utils.metrics import COMPLEX

class ORM(ABC):
    def __init__(self, class_name, json_data) -> None:
        super().__init__()
        self.class_name = class_name

        if type(json_data) == dict:
            self.deserialize(json_data)
        else:
            self.deserialize({})

    def serialize(self, relaxed) -> dict:
        if relaxed:
            return {key: self.data[key] for key in self.keys(relaxed=relaxed, dict_included=False)}

        return self.data

    def _inject_empty_values(self, value):
        if value == dict:
            return {}
        elif value == list:
            return []
        else:
            return "\U0010FFFF"

    def validate(self, json_dict) -> bool:
        json_dict = json_dict or {}

        for key, value in self.keys(relaxed=False, dict_included=True).items():
            if key not in json_dict:
                json_dict[key] = self._inject_empty_values(value)
                continue

            if not (value == dict or value == list):
                json_dict[key] = str(json_dict[key])
                continue

            if type(json_dict[key]) != value:
                json_dict[key] = self._inject_empty_values(value)

        # Remove all keys that are not in the schema. Avoid Hallucination from the model
        json_dict = {key: json_dict[key] for key in json_dict.keys(
        ) if key in self.keys(relaxed=False, dict_included=True)}

        return json_dict

    def keys(self, relaxed, dict_included, schema=None):
        if schema is None:
            schema = SCHEMA_DICT[self.class_name]

        if isinstance(schema, list):
            return {key: type(schema[0][key]) for key in schema[0].keys()}

        return {key: type(schema[key]) for key in schema.keys()}

    def deserialize(self, json_dict):
        self.data = self.validate(json_dict)

    @staticmethod
    def _evaluate_none_case(reference_keys):
        print("Evaluating None case")

        results = {}

        for key in reference_keys:
            results[key] = 0

        return results

    @staticmethod
    def _generic_evaluation(prediction, reference, relaxed):
        results = {}

        if reference is None:
            raise EvaluationException("Reference cannot be None")

        if prediction is None:
            return ORM._evaluate_none_case(reference.keys(relaxed=relaxed, dict_included=False), relaxed)

        if prediction.keys(relaxed=relaxed, dict_included=False) != reference.keys(relaxed=relaxed, dict_included=False):
            raise ORMException(
                "The keys of the prediction and reference must be the same")

        for key in prediction.keys(relaxed=relaxed, dict_included=False):

            if relaxed:
                from cv_parsing.utils.metrics import COMPLEX as metric
            else:
                from cv_parsing.utils.metrics import EXACT_MATCH as metric

            results[key] = {
                'value': 0,
                'no_prediction': 0
            }

            if prediction.data[key] == '\U0010ffff':
                results[key]['no_prediction'] = 1
                continue

            results[key]['value'] = metric.compute(
                prediction.data[key], reference.data[key])

            if results[key]['value'] == 0 and relaxed is True:
                logging.info(
                    f"Evaluation Error : Key: {key} - Prediction: {prediction.data[key]} - Reference: {reference.data[key]}")

        return results

    @staticmethod
    def _align_for_evaluation(predictions: list, references: list, keys_to_align: list):
        aligned_predictions = []
        aligned_references = []

        for prediction in predictions:
            str_concat_prediction = ' '.join(
                prediction.data[key] for key in keys_to_align)

            tmp_references = [' '.join(
                reference.data[key] for key in keys_to_align) for reference in references]

            tmp_predictions = [str_concat_prediction] * len(tmp_references)

            results = COMPLEX.compute(
                tmp_predictions, tmp_references, use_aggregator=False)

            max_score_index = results.index(max(results))

            aligned_predictions.append(prediction)
            aligned_references.append(references[max_score_index])

        return aligned_predictions, aligned_references

    @staticmethod
    def _average_dict_results(results: list, keys_to_iterate):

        if len(results) == 0:
            return {}

        accumulator = {key: {'value': 0, 'no_prediction': 0, 'hallucination': 0}
                       for key in keys_to_iterate}

        for result in results:
            for key in keys_to_iterate:
                if key not in result:
                    continue

                accumulator[key]['value'] += result[key]['value']
                accumulator[key]['no_prediction'] += result[key]['no_prediction']
                accumulator[key]['hallucination'] += 1 if (
                    result[key]['value'] == 0 and result[key]['no_prediction'] == 0) else 0

        for key in keys_to_iterate:
            accumulator[key]['value'] = accumulator[key]['value'] / \
                len(results)

            accumulator[key]['no_prediction'] = accumulator[key]['no_prediction'] / \
                len(results)

            accumulator[key]['hallucination'] = accumulator[key]['hallucination'] / \
                len(results)

        return accumulator

    @staticmethod
    def _enforce_same_length(predictions: list, references: list, class_type):
        max_length = max(len(predictions), len(references))

        if len(predictions) < max_length:
            predictions += [class_type(None)] * (max_length - len(predictions))

        elif len(references) < max_length:
            references += [class_type(None)] * (max_length - len(references))

        return predictions, references

    @staticmethod
    def _evaluate(predictions: list, references: list, relaxed: bool, keys_to_align: list, class_type):
        if len(predictions) != len(references):
            predictions, references = ORM._enforce_same_length(
                predictions, references, class_type)

        if len(keys_to_align) > 0:
            predictions, references = ORM._align_for_evaluation(
                predictions, references, keys_to_align=keys_to_align)

        if len(predictions) == 0 and len(references) == 0:
            return {}

        results = []
        no_suitable_reference = 0

        keys_to_iterate = references[0].keys(
            relaxed=relaxed, dict_included=False)

        for prediction, reference in zip(predictions, references):
            try:
                results.append(ORM._generic_evaluation(
                    prediction, reference, relaxed))
            except EvaluationException as e:
                logging.error(e)
                no_suitable_reference += 1

        return ORM._average_dict_results(results, keys_to_iterate=keys_to_iterate)

    def __str__(self) -> str:
        return str(self.serialize())



from cv_parsing.orms.ORM import ORM
from cv_parsing.utils.json_utils import SCHEMA_DICT

class LanguageLevel(ORM):
    def __init__(self, json_data) -> None:
        super().__init__("language_level", json_data=json_data)

    def keys(self, relaxed, dict_included):
        return super().keys(relaxed, dict_included, schema=SCHEMA_DICT['languages'][0]['language_level'])

    def evaluate(predictions: list, references: list, relaxed):
        return ORM._evaluate(predictions, references, relaxed,
                             keys_to_align=[], class_type=LanguageLevel)



import evaluate
from Levenshtein import ratio

rouge = evaluate.load('rouge')

NO_STR_ESCAPE = '\U0010FFFF'


class EXACT_MATCH:
    @staticmethod
    def compute(prediction, reference):
        prediction = prediction or ''
        reference = reference or ''

        if isinstance(prediction, int):
            prediction = str(prediction)

        if isinstance(reference, int):
            reference = str(reference)

        return float(prediction.strip().upper() == reference.strip().upper())

class COMPLEX:

    @staticmethod
    def _compute_rouge(predictions, references):
        return rouge.compute(predictions=predictions, references=references, use_aggregator=False)['rougeL']

    @staticmethod
    def _compute_levenshtein(prediction, reference):
        return ratio(prediction, reference)

    @staticmethod
    def compute(predictions, references, use_aggregator=True):
        if not isinstance(predictions, list):
            predictions = [predictions]

        if not isinstance(references, list):
            references = [references]

        results = []
        
        rouge_calcs = {
            'predictions': [],
            'references': []
        }

        lev_calcs = []

        for prediction, reference in zip(predictions, references):
            
            prediction = prediction.strip().upper().encode('ascii', 'ignore').decode('ascii')
            reference = reference.strip().upper().encode('ascii', 'ignore').decode('ascii')

            # Single Token is not suitable for Rouge Longest Common Subsequence
            if len(prediction.split()) == 1 and len(reference.split()) == 1:
                lev_calcs.append((prediction, reference))
            else:
                rouge_calcs['predictions'].append(prediction)
                rouge_calcs['references'].append(reference)

        for prediction, reference in lev_calcs:
            results.append(COMPLEX._compute_levenshtein(prediction, reference))
        
        if len(rouge_calcs['predictions']) > 0:
            results.extend(COMPLEX._compute_rouge(
                predictions=rouge_calcs['predictions'], references=rouge_calcs['references']))

        if not use_aggregator:
            return results

        return round(sum(results) / len(results), 3)



import os
from openai import OpenAI
from cv_parsing.parser.ParsingStrategy import ParsingStrategy
from cv_parsing.orms.Person import Person
from cv_parsing.exceptions.PromptException import PromptException
from cv_parsing.utils.json_utils import repair_json

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = CURRENT_DIR.split("src")[0]
DATA_DIR = os.path.join(ROOT_DIR, "data")


class OpenAIStrategy(ParsingStrategy):
    def __init__(self, prompts: list, model=None, api_key=None, temperature=None) -> None:
        super().__init__("OpenAI")

        self.prompts = prompts

        self.model = model or self.env('OPENAI_MODEL')

        self.api_key = api_key or self.env('OPENAI_API_KEY')

        self.temperature = temperature or self.env('MODEL_TEMPERATURE')

        print(
            f"Executing OpenAIStrategy with model: {self.model}, api_key: {self.api_key}, temperature: {self.temperature}")

        if not self.model:
            raise ValueError("Model is required")

        if not self.api_key:
            raise ValueError("API Key is required")

        if not self.temperature:
            raise ValueError("Temperature is required")

        self.openai = OpenAI(api_key=self.api_key)

    def _validate_model_output(self, gpt_response) -> bool:
        gpt_response = gpt_response.to_dict()

        if 'choices' not in gpt_response or len(gpt_response['choices']) == 0:
            raise PromptException("Invalid GPT Response")

        if 'message' not in gpt_response['choices'][0]:
            raise PromptException("Invalid GPT Response")

        if 'content' not in gpt_response['choices'][0]['message']:
            raise PromptException("Invalid GPT Response")

        try:
            return repair_json(
                gpt_response['choices'][0]['message']['content'])
        except Exception as e:
            raise PromptException("Invalid GPT Response")

    def parse(self, normalized_text: str) -> Person:
        dict_results = {}

        for prompt in self.prompts:
            response = self.openai.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": prompt.header},
                    {"role": "user", "content": prompt.substitute(
                        cv=normalized_text)},
                ],
                temperature=float(self.temperature),
                response_format={"type": "json_object"},
            )

            dict_results.update(self._validate_model_output(response))

        return Person(dict_results)

    @staticmethod
    def models():
        with open(os.path.join(DATA_DIR, "models.jsonl"), 'r', encoding='utf-8') as f:
            return repair_json(f.read())["OpenAI"]



from cv_parsing.orms.ORM import ORM
from cv_parsing.utils.json_utils import SCHEMA_DICT
# Other ORMs
from cv_parsing.orms.PersonalInformation import PersonalInformation
from cv_parsing.orms.Job import Job
from cv_parsing.orms.Education import Education
from cv_parsing.orms.Language import Language
from cv_parsing.exceptions.EvaluationException import EvaluationException
from cv_parsing.utils.logger import logging
from datasets import Dataset


class Person(ORM):
    def __init__(self, json_data) -> None:
        super().__init__(class_name="person", json_data=json_data)

    def keys(self, relaxed, dict_included) -> dict:
        results = {}

        if not dict_included:
            return results

        for key in SCHEMA_DICT.keys():
            results[key] = type(SCHEMA_DICT[key])

        return results

    def deserialize(self, json_dict):
        self.data = self.validate(json_dict)

        self.personal_information = PersonalInformation(
            self.data['personal_information'])

        self.education = [Education(education)
                          for education in self.data['education']]

        self.jobs = [Job(job) for job in self.data['jobs']]

        self.languages = [Language(language)
                          for language in self.data['languages']]

    def serialize(self, relaxed=False) -> dict:
        results = super().serialize(relaxed)

        results['personal_information'] = self.personal_information.serialize(
            relaxed)

        results['education'] = [education.serialize(relaxed)
                                for education in self.education]

        results['jobs'] = [job.serialize(relaxed) for job in self.jobs]

        results['languages'] = [language.serialize(relaxed)
                                for language in self.languages]

        return results
        # TODO: Adicional ORMs must be added here

    @staticmethod
    def _align_for_evaluation(predictions, references):
        aligned_predictions = []
        aligned_references = []

        for reference in references:
            filename = reference['filename']
            # Find the prediction with the same filename
            prediction = next(
                (p for p in predictions if p['filename'].split('\\')[-1] == filename), None)

            aligned_predictions.append(Person(prediction))
            aligned_references.append(Person(reference))

        return aligned_predictions, aligned_references

    @staticmethod
    def evaluate(predictions: Dataset, references: Dataset, relaxed):
        results = {}

        predictions, references = Person._align_for_evaluation(
            predictions, references)

        if len(predictions) != len(references):
            raise EvaluationException(
                "The number of predictions and references must be the same")

        # results = ORM._generic_evaluation(predictions, references, relaxed)
        results.update({'personal_information': PersonalInformation.evaluate(
            [prediction.personal_information for prediction in predictions], [reference.personal_information for reference in references], relaxed)})

        print("Evaluated Personal Information")

        results.update({'education': ORM._average_dict_results([Education.evaluate(
            prediction.education, reference.education, relaxed) for prediction, reference in zip(predictions, references)], keys_to_iterate=references[0].education[0].keys(relaxed, dict_included=False))})

        print("Evaluated Education")

        """
        """
        results.update({'jobs': ORM._average_dict_results([Job.evaluate(
            prediction.jobs, reference.jobs, relaxed) for prediction, reference in zip(predictions, references)], keys_to_iterate=references[0].jobs[0].keys(relaxed, dict_included=False))})

        print("Evaluated Jobs")
        languages = [Language.evaluate(
            prediction.languages, reference.languages, relaxed) for prediction, reference in zip(predictions, references)]

        results.update({'languages': Language._average_dict_results(
            languages, keys_to_iterate=references[0].languages[0].keys(relaxed, dict_included=False))})
        """
        """
        print("Evaluated Languages")

        print("Evaluation Completed")

        # TODO: Adicional ORMs must be added here

        return results



import os
from abc import ABC, abstractmethod
from environs import Env
from cv_parsing.utils.json_utils import repair_json

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = CURRENT_DIR.split("src")[0]
DATA_DIR = os.path.join(ROOT_DIR, "data")


class ParsingStrategy(ABC):
    def __init__(self, name, *args, **kwargs) -> None:
        self._name = name
        self.env = Env()
        self.env.read_env()

    @abstractmethod
    def _validate_model_output(self, text: str) -> dict:
        pass

    @abstractmethod
    # def parse(self, raw_pdf_text: str) -> Person:
    def parse(self, raw_pdf_text: str):
        pass

    @staticmethod
    @abstractmethod
    def models():
        pass

    @property
    def name(self):
        return self._name



import os
import logging
import time
import tempfile


# Get temporary directory

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = CURRENT_DIR.split('src')[0]
OUT_DIR = os.path.join(ROOT_DIR, 'out')

timestamp = time.strftime("%Y%m%d-%H%M%S")


def load_logger(path=os.path.join(OUT_DIR, f'cv_{timestamp}.log')):
    logging.basicConfig(level=logging.INFO, filename=path,
                        filemode='w', format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


if os.path.exists(OUT_DIR):
    load_logger()
else:
    fp = tempfile.NamedTemporaryFile(delete=False)
    load_logger(fp.name)



from cv_parsing.prompts.Prompt import Prompt
from string import Template

prompt_header = """
    Tu és um assistente para a equipa de Recursos Humanos de uma empresa portuguesa.
    Foste encarregado de extrair informações de um currículo vitae.
    A tua resposta deve ignorar qualquer texto introdutório e focar-se nas informações estruturadas em json.
    A maioria dos candidatos está a candidatar-se a um emprego em Português, mas por vezes os currículos estão em Inglês.
"""


class BasicPrompt(Prompt):
    def __init__(self, feature="all", header=prompt_header) -> None:
        super().__init__('portuguese', feature, header,  Template("""
    A String não estruturada que representa o currículo é a seguinte:

    $cv

    Deves extrair as seguintes informações do currículo:
    
    $json_schema  

    Algumas das informações podem não estar presentes no currículo. Nesses casos, deve produzir um valor json null.

    Os empregos podem ser na indústria ou académicos. Se for um emprego académico, é muito provável que o empregador seja uma universidade ou centro de investigação.

    Se precisares de efetuar quaisquer cálculos relativos a datas: 
     - pode assumir que o ano atual é $current_year.
     - se a data se estender até ao presente, deve considerar a data como "PRESENTE".
     
    Se for fornecida a idade, invés da data de nascimento, deve fornecer apenas o ano de nascimento. 
     - Qualquer calculo considera que o ano atual é $current_year.

    No caso das línguas: 
     - Deve considerar três níveis: A, B, C.
     - A língua materna deve ser considerada como C

    A informação deve ser devolvida num json válido.
    """))

# TODO: Implement Multi-Prompt Chaining Reasoning


class ExtractPersonalInformationPrompt(Prompt):
    pass


class ExtractJobsPrompt(Prompt):
    pass


class ExtractEducationPrompt(Prompt):
    pass


class ExtractLanguagesPrompt(Prompt):
    pass



from cv_parsing.orms.ORM import ORM
from cv_parsing.exceptions.ORMException import ORMException
from cv_parsing.orms.LanguageLevel import LanguageLevel


class Language(ORM):
    def __init__(self, json_data) -> None:
        super().__init__(class_name="languages", json_data=json_data)

    def keys(self, relaxed, dict_included):
        results = super().keys(relaxed, dict_included)

        if not dict_included:
            results.pop('language_level')

        return results

    @staticmethod
    def _average_dict_results(results: list, keys_to_iterate):
        accumulator = ORM._average_dict_results(results, keys_to_iterate)

        accumulator.update(
            ORM._average_dict_results([language_level for language_level in results], [
                                      "speaking", "reading", "writing"])
        )

        return accumulator

    @staticmethod
    def evaluate(predictions: list, references: list, relaxed):

        results = ORM._evaluate(
            predictions, references, relaxed, keys_to_align=["name"], class_type=Language)

        predictions, references = ORM._align_for_evaluation(
            predictions, references, ["name"])

        accumulator = [LanguageLevel.evaluate(
            [prediction.language_level], [reference.language_level], relaxed) for prediction, reference in zip(predictions, references)]

        results.update(ORM._average_dict_results(
            accumulator, keys_to_iterate=["speaking", "reading", "writing"]))

        return results

    def deserialize(self, json_dict):
        self.data = self.validate(json_dict)

        self.language = self.data['name']
        self.language_level = LanguageLevel(self.data['language_level'])

    def serialize(self, relaxed=False) -> dict:
        results = super().serialize(relaxed)

        results['language_level'] = self.language_level.serialize(relaxed)

        return results



from cv_parsing.parser.ParsingStrategy import ParsingStrategy
from cv_parsing.utils.clean_text import clean_text
from cv_parsing.orms.Person import Person


class Parser:
    def __init__(self, parsing_strategy: ParsingStrategy) -> None:
        self.parsing_strategy = parsing_strategy

    def parse(self, raw_pdf_text: str) -> Person:   
        return self.parsing_strategy.parse(clean_text(raw_pdf_text))



from cv_parsing.prompts.Prompt import Prompt
from string import Template
from cv_parsing.utils.json_utils import SCHEMA_DICT, JSON_STR


prompt_header = """
    You are an assistant for the Human Resource Team at a Portuguese employer. 
    You have been tasked with extracting information from a resume.
    Your answer should be a single valid json response.
    Your answer must skip any introductory/trailing text and focus on the json response.
    Most of the candidates are applying for a job in Portugal, so most of the resumes are in Portuguese.
    This is an extractive task, all the information required is present in the resume.
"""


class BasicPrompt(Prompt):
    def __init__(self, feature="all", schema=JSON_STR, header=prompt_header) -> None:
        super().__init__('english', feature, schema, header, Template(
            f"""
    The unstructured String that represents the resume is as follows:

    $cv

    You should extract the following information from the resume:
    
    $json_schema   

    Some of the information may not be present in the resume. In such cases, you should produce null.

    The jobs can either be a job in industry or academic

    If you need to perform any calculations regarding dates: 
     - you can assume that the current year is $current_year.
     - if the date extends to the present, you should consider the date as "PRESENT".
     
    If the age is provided: 
     - instead of the birth_year, you should provide the birthyear only.
     - The age should be calculated based on the current year $current_year.

    In the cases of the languages: 
     - You should consider Four language_levels: A, B, C, F.
     - The mother tongue should be considered an F.
    """))


class ExtractPersonalInformationPrompt(Prompt):
    def __init__(self, feature="personal_information", schema=SCHEMA_DICT['personal_information'], header=prompt_header) -> None:
        super().__init__('english', feature, schema, header, Template(
            f"""
    The unstructured String that represents the resume is as follows:

    $cv

    You should extract the following information from the resume:
    
    {{"{feature}":$json_schema}}
    
    Some of the information may not be present in the resume. In such cases, you should produce null.

    If you need to perform any calculations regarding dates: 
     - you can assume that the current year is $current_year.
        
    """))


class ExtractJobsPrompt(Prompt):
    def __init__(self, feature="jobs", schema=SCHEMA_DICT['jobs'], header=prompt_header) -> None:
        super().__init__('english', feature, schema, header, Template(
            f"""
  
    The unstructured String that represents the resume is as follows:

    $cv

    Your job is to extract information regarding the jobs that the candidate has had.
    
    You should extract the following information from the resume:
    
    {{"{feature}":$json_schema}}
        
    Some of the information may not be present in the resume. In such cases, you should produce null.

    The jobs can either be in industry or academia:
     - If it is an academic job, it is heavily like the employer to be a university or research center     
    
     If you need to perform any calculations regarding dates: 
     - you can assume that the current year is $current_year.
     - if the date extends to the present, you should consider the date as "PRESENT".

    Avoid providing additional information rather than the json response. No computer code or comments should be present in the response.
    """))


class ExtractEducationPrompt(Prompt):
    def __init__(self, feature="education", schema=SCHEMA_DICT['education'], header=prompt_header) -> None:
        super().__init__('english', feature, schema, header, Template(
            f"""
    The unstructured String that represents the resume is as follows:

    $cv

    Your job is to extract information regarding the education that the candidate has.

    You should extract the following information from the resume:
    
    {{"{feature}":$json_schema}}

    Some of the information may not be present in the resume. In such cases, you should produce null.

    If you need to perform any calculations regarding dates: 
     - you can assume that the current year is $current_year.
     - if the date extends to the present, you should consider the date as "PRESENT".        
    """))


class ExtractLanguagesPrompt(Prompt):
    def __init__(self, feature="languages", schema=SCHEMA_DICT['languages'], header=prompt_header) -> None:
        super().__init__('english', feature, schema, header, Template(
            f"""
    
    The unstructured String that represents the resume is as follows:

    $cv

    Your job is to extract information regarding the language that the candidate knows how to speak.

    You should extract the following information from the resume:
    
    {{"{feature}":$json_schema}}

    Some of the information may not be present in the resume. In such cases, you should produce null.
    
    You should consider four proficiency language_levels: 
     -A
     -B 
     -C
     -F


    If there is mention of mother tongue / native should be considered F.
    """))



import time
from string import Template
from cv_parsing.utils.clean_text import clean_text


class Prompt:
    def __init__(self, language, feature, schema, header, template: Template, current_year=time.localtime().tm_year) -> None:
        self.name = self.__class__.__name__
        self.language = language
        self.feature = feature
        self.schema = schema
        self.header = header
        self.template = template
        self.current_year = current_year

    def substitute(self, **kwargs):
        return f"""
        {self.header}
        {self.template.safe_substitute(
            {**kwargs, 'json_schema': clean_text(self.schema), 'current_year': clean_text(self.current_year)})}
        """



import ollama
import os
from cv_parsing.parser.ParsingStrategy import ParsingStrategy
from cv_parsing.orms.Person import Person
from cv_parsing.utils.json_utils import repair_json
from cv_parsing.exceptions.PromptException import PromptException

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = CURRENT_DIR.split("src")[0]
DATA_DIR = os.path.join(ROOT_DIR, "data")

class OllamaStrategy(ParsingStrategy):
    def __init__(self, prompts: list, model=None, temperature=None) -> None:
        super().__init__("Ollama")

        self.prompts = prompts

        self.model = model or self.env('OLLAMA_MODEL')

        ollama.pull(self.model)

        self.temperature = temperature or self.env('MODEL_TEMPERATURE')

        print(
            f"Executing OllamaStrategy with model: {self.model}, temperature: {self.temperature}")

        if not self.model:
            raise ValueError("Model is required")

        if not self.temperature:
            raise ValueError("Temperature is required")

    def _validate_model_output(self, ollama_response):
        if 'message' not in ollama_response:
            raise PromptException("Invalid Ollama Response")

        if 'content' not in ollama_response['message']:
            raise PromptException("Invalid Ollama Response")

        try:
            return repair_json(
                ollama_response['message']['content'])
        except Exception as e:
            raise PromptException("Invalid Ollama Response")

    def parse(self, normalized_text: str) -> Person:
        dict_results = {}

        for prompt in self.prompts:
            response = ollama.chat(
                model=self.model,
                messages=[
                    {"role": "system", "content": prompt.header},
                    {"role": "user", "content": prompt.substitute(
                        cv=normalized_text)},
                ],
                options={"temperature": self.temperature},
                format='json',
            )
            
            dict_results.update(self._validate_model_output(response))

        return Person(dict_results)
    
    @staticmethod
    def models():
        with open(os.path.join(DATA_DIR, "models.jsonl"), 'r', encoding='utf-8') as f:
            return repair_json(f.read())["Ollama"]


from environs import Env


def load_env():
    env = Env()
    env.read_env()

    return env


def load_open_ai_key():
    env = load_env()
    open_ai_key = env('OPEN_AI_KEY', None)

    if open_ai_key is None:
        raise ValueError('OPEN_AI_KEY is not set in the environment')

    return open_ai_key



from cv_parsing.orms.Person import Person
import plotly.express as px
from plotly.subplots import make_subplots
import pandas as pd


class Evaluation:
    @staticmethod
    def evaluate(predictions_dataset, references_dataset):

        results = {
            "relaxed": Person.evaluate(predictions_dataset, references_dataset, relaxed=True),
            "strict": Person.evaluate(predictions_dataset, references_dataset, relaxed=False),
        }

        return results

    @staticmethod
    def _plot_results(results_relaxed, results_strict, title):
        # Ensure all relaxed keys are present in strict keys
        if not all(key in results_strict for key in results_relaxed):
            raise ValueError(
                "All Relaxed evaluations must be present in Strict evaluations")

        data = {
            'property': [],
            'evaluation': [],
            'value': []
        }

        for key in list(results_strict.keys()):
            data['property'].append(key)
            data['evaluation'].append("Strict")
            data['value'].append(round(results_strict[key]['value'], 3))

        for key in list(results_relaxed.keys()):
            data['property'].append(key)
            data['evaluation'].append("Relaxed")
            data['value'].append(round(results_relaxed[key]['value'], 3))

        for key in list(results_relaxed.keys()):
            data['property'].append(key)
            data['evaluation'].append("hallucination")
            data['value'].append(
                round(results_relaxed[key]['hallucination'], 3))

        for key in list(results_relaxed.keys()):
            data['property'].append(key)
            data['evaluation'].append("no_prediction")
            data['value'].append(
                round(results_relaxed[key]['no_prediction'], 3))

        return px.bar(pd.DataFrame(data),
                      x="value",
                      y="property",
                      color='evaluation',
                      barmode='group',
                      orientation='h',
                      text_auto=True,
                      title=title)

    @staticmethod
    def plot_results(results, title):
        fig = make_subplots(rows=2, cols=2, subplot_titles=[
            "Personal Information",
            "Jobs",
            "Education",
            "Languages",
        ])

        row = 1

        for i, key in enumerate(["personal_information", "jobs", "education", "languages"]):

            if i % 2 == 0 and i > 0:
                row += 1

            bar = Evaluation._plot_results(
                results["relaxed"][key],
                results["strict"][key],
                key)

            for trace in bar.data:
                fig.add_trace(trace, row=row, col=i % 2 + 1)

        fig.update_layout(font=dict(size=18))
        fig.update_xaxes(range=[0, 1])
        fig.update_layout(title_text=title)
        # Enforce floating point precision

        return fig

    @staticmethod
    def compare_models(results_filenames):
        pass



import os
import json
from cv_parsing.data.Dataloader import Dataloader
from datasets import Dataset
from cv_parsing.exceptions.PromptException import PromptException
from tqdm import tqdm

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = CURRENT_DIR.split('src')[0]
OUT_DIR = os.path.join(ROOT_DIR, 'out')



class Predict:
    def __init__(self, parser, debug=False, dataloader=Dataloader()):
        self.parser = parser
        self.debug = debug
        self.dataloader = dataloader

    def run_file(self, filepath):
        raw_pdf_text = self.dataloader.load_pdf(filepath)

        person = self.parser.parse(raw_pdf_text)

        person.data['filename'] = filepath.split('/')[-1]

        return person

    def all_files(self):
        persons = []

        for filepath in tqdm(self.dataloader.load_all_pdfs()):

            try:
                person = self.run_file(filepath)
            except PromptException as e:
                print(f"Error Model Response: {filepath} | Skipping...")
                continue
            except OverflowError as e:
                print(f"Overflow Error: {filepath} | Skipping...")
                continue

            persons.append(person)

        return persons

    def create_dataset(self, config_name):
        persons = self.all_files()

        persons = [person.serialize()
                   for person in persons if person is not None]

        config_name = config_name.replace(':', '-')

        with open(os.path.join(OUT_DIR, f"results-{config_name}.json"), 'w+', encoding='utf-8') as f:
            json.dump(persons, f, indent=4, ensure_ascii=False)

        try:
            dataset = Dataset.from_list(persons, split='train')

            print(
                f"Pusing to Hub: innovpoint/curriculum-vitae with config_name: {config_name} and length: {len(dataset)}")

            dataset.push_to_hub('innovpoint/curriculum-vitae',
                                config_name=config_name, private=True)
        except Exception as e:
            print(f"Error pushing to hub: {e}")
